"""
FIXME: 感觉这个 component 设计有点问题，做的事情太多，得拆！
"""

from PyQt5.QtCore import (
    QAbstractListModel,
    QModelIndex,
    QSize,
    Qt,
    QVariant,
)
from PyQt5.QtWidgets import (
    QListView,
    QSizePolicy,
)


class ProviderModel(object):
    """
    一个 ProviderModel 对应左边栏「我的音乐」下的一个项目

    FIXME: 或许不应该命名为 ProviderModel，而是
    """
    def __init__(self, name, icon='♬ ', desc='', on_click=None, **kwargs):
        self.name = name
        self.icon = icon
        self.desc = desc
        self._on_click = on_click


class ProvidersModel(QAbstractListModel):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._provider_list = []
        self._association = {}

    def assoc(self, provider_id, pm):
        self._association[provider_id] = pm
        self._provider_list.append(provider_id)
        self.insertRow(len(self._provider_list))

    def get(self, provider_id):
        return self._association.get(provider_id)

    def remove(self, provider_id):
        if not self._association.pop(provider_id, None):
            row = self._provider_list.index(provider_id)
            self._provider_list.remove(provider_id)
            self.removeRow(row)

    def rowCount(self, parent=QModelIndex()):
        return len(self._provider_list)

    def flags(self, index):
        return Qt.ItemIsSelectable | Qt.ItemIsEnabled

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return QVariant()
        row = index.row()
        if row >= len(self._provider_list) or row < 0:
            return QVariant()

        provider_id = self._provider_list[row]
        provider = self._association[provider_id]
        if role == Qt.DisplayRole:
            return provider.icon + ' ' + provider.name
        elif role == Qt.ToolTipRole:
            return provider.desc
        elif role == Qt.UserRole:
            return provider
        return QVariant()


class ProvidersView(QListView):
    def __init__(self, parent):
        super().__init__(parent)

        self.setMinimumHeight(80)
        self.setAttribute(Qt.WA_MacShowFocusRect, 0)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)

        self.clicked.connect(self._on_clicked)

    def _on_clicked(self, index):
        provider = index.data(role=Qt.UserRole)
        if provider._on_click:
            provider._on_click()

    def sizeHint(self):
        height = 10
        if self.model().rowCount() > 0:
            height = self.model().rowCount() * self.sizeHintForRow(0)
        return QSize(self.width(), height)



import asyncio
import logging
import sys

from PyQt5.QtCore import Qt, QTime, pyqtSignal, pyqtSlot, QTimer
from PyQt5.QtGui import QFontMetrics, QPainter, QFont, QKeySequence
from PyQt5.QtWidgets import (
    QAction,
    QComboBox,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QMenu,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QSlider,
    QSplitter,
    QStatusBar,
    QStyle,
    QVBoxLayout,
    QWidget,
)

from feeluown.components.separator import Separator
from feeluown.components.playlists import (
    PlaylistsView,
    PlaylistsModel,
)
from feeluown.components.provider import ProvidersView
from feeluown.components.history import HistoriesView
from feeluown.containers.magicbox import MagicBox
from feeluown.containers.table_container import SongsTableContainer

from .consts import PlaybackMode
from .helpers import use_mac_theme
from .utils import parse_ms


logger = logging.getLogger(__name__)


class VolumeSlider(QSlider):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setOrientation(Qt.Horizontal)
        #self.setRange(0, 100)   # player volume range
        #self.setValue(100)
        self.setToolTip('调教播放器音量')


class ProgressSlider(QSlider):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setOrientation(Qt.Horizontal)

    def set_duration(self, ms):
        self.setRange(0, ms / 1000)

    def update_state(self, ms):
        self.setValue(ms / 1000)


class PlayerControlPanel(QFrame):

    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        class Button(QPushButton):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                # 按钮文字一般是一个 symbol，长度控制为 40 是满足需求的
                self.setMaximumWidth(40)

        # initialize sub widgets
        self._layout = QHBoxLayout(self)
        self.previous_btn = Button(self)
        self.pp_btn = Button(self)
        self.next_btn = Button(self)
        self.pms_btn = Button(self)
        self.volume_btn = Button(self)
        self.playlist_btn = Button(parent=self)

        self.previous_btn.setObjectName('previous_btn')
        self.pp_btn.setObjectName('pp_btn')
        self.next_btn.setObjectName('next_btn')
        self.playlist_btn.setObjectName('playlist_btn')

        self.progress_slider = ProgressSlider(self)

        # TODO(simple): implementation
        self.pms_btn.setToolTip('修改播放模式（未实现，欢迎 PR）')
        self.volume_btn.setToolTip('调整音量（未实现，欢迎 PR）')
        self.playlist_btn.setToolTip('显示当前播放列表')
        self.progress_slider.setToolTip('拖动调节进度（未实现，欢迎 PR）')

        if not use_mac_theme():
            self.previous_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaSkipBackward))
            self.pp_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
            self.next_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaSkipForward))
            self.volume_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaVolume))
            self.playlist_btn.setText('🎶')
        else:
            self.pp_btn.setCheckable(True)

        self.song_title_label = QLabel('No song is playing.', parent=self)
        self.song_title_label.setAlignment(Qt.AlignCenter)
        self.duration_label = QLabel('00:00', parent=self)
        self.position_label = QLabel('00:00', parent=self)

        self.next_btn.clicked.connect(self._app.player.play_next)
        self.previous_btn.clicked.connect(self._app.player.play_previous)
        self.pp_btn.clicked.connect(self._app.player.toggle)

        # set widget layout
        self.progress_slider.setMinimumWidth(480)
        self.progress_slider.setMaximumWidth(600)
        self.progress_slider.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self._sub_layout = QVBoxLayout()
        self._sub_layout.addWidget(self.song_title_label)
        self._sub_layout.addWidget(self.progress_slider)

        self._layout.addSpacing(15)
        self._layout.addWidget(self.previous_btn)
        self._layout.addSpacing(5)
        self._layout.addWidget(self.pp_btn)
        self._layout.addSpacing(5)
        self._layout.addWidget(self.next_btn)
        self._layout.addSpacing(20)
        self._layout.addWidget(self.volume_btn)
        self._layout.addStretch(0)
        self._layout.addWidget(self.position_label)
        self._layout.addSpacing(7)
        self._layout.addLayout(self._sub_layout)
        self._layout.addSpacing(7)
        self._layout.addWidget(self.duration_label)
        self._layout.addSpacing(5)
        self._layout.addStretch(0)
        self._layout.addWidget(self.pms_btn)
        self._layout.addWidget(self.playlist_btn)
        self._layout.addSpacing(10)

        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)

    def on_duration_changed(self, duration):
        m, s = parse_ms(duration)
        t = QTime(0, m, s)
        self.duration_label.setText(t.toString('mm:ss'))

    def on_position_changed(self, position):
        m, s = parse_ms(position)
        t = QTime(0, m, s)
        self.position_label.setText(t.toString('mm:ss'))

    def on_playback_mode_changed(self, playback_mode):
        self.pms_btn.setText(playback_mode.value)

    def on_player_song_changed(self, song):
        self.song_title_label.setText(
            '♪  {title} - {artists_name}'.format(
                title=song.title,
                artists_name=song.artists_name))


class TopPanel(QFrame):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self._layout = QHBoxLayout(self)
        self.pc_panel = PlayerControlPanel(self._app, self)
        self.setObjectName('top_panel')

        self.setFixedHeight(60)

        self._layout.addWidget(self.pc_panel)


class LeftPanel(QFrame):

    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self.library_header = QLabel('我的音乐', self)
        self.playlists_header = QLabel('歌单列表', self)
        self.history_header = QLabel('浏览历史记录', self)

        class Container(QWidget):
            def __init__(self, label, view, parent=None):
                super().__init__(parent)

                self._layout = QVBoxLayout(self)
                self._layout.setContentsMargins(0, 0, 0, 0)
                self._layout.setSpacing(0)
                label.setFixedHeight(25)
                self._layout.addWidget(label)
                self._layout.addWidget(view)

        self.playlists_view = PlaylistsView(self)
        self.providers_view = ProvidersView(self)
        self.histories_view = HistoriesView(self)

        self._providers_con = Container(self.library_header, self.providers_view)
        self._histories_con = Container(self.history_header, self.histories_view)
        self._playlists_con = Container(self.playlists_header, self.playlists_view)

        self._splitter = QSplitter(Qt.Vertical, self)

        self.providers_view.setModel(self._app.providers)
        self.histories_view.setModel(self._app.histories)
        self.playlists_view.setModel(self._app.playlists)

        self._layout = QVBoxLayout(self)
        if use_mac_theme():
            self._layout.setSpacing(0)
            self._layout.setContentsMargins(6, 4, 0, 0)
        self._splitter.addWidget(self._providers_con)
        self._splitter.addWidget(self._histories_con)
        self._splitter.addWidget(self._playlists_con)
        self._layout.addWidget(self._splitter)

        self.providers_view.setFrameShape(QFrame.NoFrame)
        self.playlists_view.setFrameShape(QFrame.NoFrame)
        self.histories_view.setFrameShape(QFrame.NoFrame)
        self.setMinimumWidth(180)
        self.setMaximumWidth(250)

        self.playlists_view.show_playlist.connect(
            lambda pl: asyncio.ensure_future(self.show_model(pl)))
        self.histories_view.show_model.connect(
            lambda model: asyncio.ensure_future(self.show_model(model)))

    async def show_model(self, playlist):
        await self._app.ui.table_container.show_model(playlist)


class RightPanel(QFrame):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self._layout = QHBoxLayout(self)
        self.table_container = SongsTableContainer(self._app, self)
        self._layout.addWidget(self.table_container)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)


class SongLabel(QLabel):
    def __init__(self, text=None, parent=None):
        super().__init__(text, parent)
        self.set_song('No song is playing')

    def set_song(self, song_text):
        self.setText('♪  ' + song_text + ' ')


class MessageLabel(QLabel):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self.setObjectName('message_label')
        self._interval = 3
        self.timer = QTimer()
        self.queue = []
        self.hide()

        self.timer.timeout.connect(self.access_message_queue)

    @property
    def common_style(self):
        style_str = '''
            #{0} {{
                padding-left: 3px;
                padding-right: 5px;
            }}
        '''.format(self.objectName())
        return style_str

    def _set_error_style(self):
        theme = self._app.theme_manager.current_theme
        style_str = '''
            #{0} {{
                background: {1};
                color: {2};
            }}
        '''.format(self.objectName(),
                   theme.color1_light.name(),
                   theme.color7_light.name())
        self.setStyleSheet(style_str + self.common_style)

    def _set_normal_style(self):
        theme = self._app.theme_manager.current_theme
        style_str = '''
            #{0} {{
                background: {1};
                color: {2};
            }}
        '''.format(self.objectName(),
                   theme.color6_light.name(),
                   theme.color7.name())
        self.setStyleSheet(style_str + self.common_style)

    def show_message(self, text, error=False):
        if self.isVisible():
            self.queue.append({'error': error, 'message': text})
            self._interval = 1.5
            return
        if error:
            self._set_error_style()
        else:
            self._set_normal_style()
        self.setText(str(len(self.queue)) + ': ' + text)
        self.show()
        self.timer.start(self._interval * 1000)

    def access_message_queue(self):
        self.hide()
        if self.queue:
            m = self.queue.pop(0)
            self.show_message(m['message'], m['error'])
        else:
            self._interval = 3


class NetworkStatus(QLabel):
    def __init__(self, app, text=None, parent=None):
        super().__init__(text, parent)
        self._app = app

        self.setToolTip('这里显示的是当前网络状态')
        self.setObjectName('network_status_label')
        self._progress = 100
        self._show_progress = False

        self.set_state(1)

    def paintEvent(self, event):
        if self._show_progress:
            painter = QPainter(self)
            p_bg_color = self._app.theme_manager.current_theme.color0
            painter.fillRect(self.rect(), p_bg_color)
            bg_color = self._app.theme_manager.current_theme.color3
            rect = self.rect()
            percent = self._progress * 1.0 / 100
            rect.setWidth(int(rect.width() * percent))
            painter.fillRect(rect, bg_color)
            painter.drawText(self.rect(), Qt.AlignVCenter | Qt.AlignHCenter,
                             str(self._progress) + '%')
            self._show_progress = False
        else:
            super().paintEvent(event)

    @property
    def common_style(self):
        theme = self._app.theme_manager.current_theme
        style_str = '''
            #{0} {{
                background: {1};
                color: {2};
                padding-left: 5px;
                padding-right: 5px;
                font-size: 14px;
                font-weight: bold;
            }}
        '''.format(self.objectName(),
                   theme.color3.name(),
                   theme.background.name())
        return style_str

    def set_theme_style(self):
        self.setStyleSheet(self.common_style)

    def _set_error_style(self):
        theme = self._app.theme_manager.current_theme
        style_str = '''
            #{0} {{
                background: {1};
            }}
        '''.format(self.objectName(),
                   theme.color5.name())
        self.setStyleSheet(self.common_style + style_str)

    def _set_normal_style(self):
        self.setStyleSheet(self.common_style)

    def set_state(self, state):
        if state == 0:
            self._set_error_style()
            self.setText('✕')
        elif state == 1:
            self._set_normal_style()
            self.setText('✓')

    def show_progress(self, progress):
        self._progress = progress
        self._show_progress = True
        if self._progress == 100:
            self._show_progress = False
        self.update()


class Ui(object):
    def __init__(self, app):
        self._app = app
        self._layout = QVBoxLayout(app)
        self._bottom_layout = QHBoxLayout()
        self._top_separator = Separator(parent=app)
        self._splitter = QSplitter(app)
        if use_mac_theme():
            self._splitter.setHandleWidth(0)

        # NOTE: 以位置命名的部件应该只用来组织界面布局，不要
        # 给其添加任何功能性的函数
        self.top_panel = TopPanel(app, app)
        self.left_panel = LeftPanel(self._app, self._splitter)
        self.right_panel = RightPanel(self._app, self._splitter)

        # alias
        self.pc_panel = self.top_panel.pc_panel
        self.table_container = self.right_panel.table_container
        self.magicbox = MagicBox(self._app)

        # 对部件进行一些 UI 层面的初始化
        self._splitter.addWidget(self.left_panel)
        self._splitter.addWidget(self.right_panel)

        self.right_panel.setMinimumWidth(780)
        self.left_panel.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        self.right_panel.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        if use_mac_theme():
            self._layout.addWidget(self.magicbox)
            self._layout.addWidget(self._splitter)
            self._layout.addWidget(self._top_separator)
            self._layout.addWidget(self.top_panel)
        else:
            self._layout.addWidget(self.top_panel)
            self._layout.addWidget(self._top_separator)
            self._layout.addWidget(self._splitter)
            self._layout.addWidget(self.magicbox)

        # self._layout.addLayout(self._bottom_layout)
        # self._bottom_layout.addWidget(self.magicbox)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.top_panel.layout().setSpacing(0)
        self.top_panel.layout().setContentsMargins(0, 0, 0, 0)

        self.pc_panel.playlist_btn.clicked.connect(self.show_player_playlist)

        self._app.hotkey_manager.registe(
            [QKeySequence('Ctrl+F'), QKeySequence(':'), QKeySequence('Alt+x')],
            self.magicbox.setFocus
        )

    def show_player_playlist(self):
        songs = self._app.playlist.list()
        self.table_container.show_player_playlist(songs)



import asyncio
import logging

from PyQt5.QtCore import QObject

from fuocore.netease.provider import provider
from fuocore.netease.models import NUserModel, search

from .login_controller import LoginController
from .ui import LoginDialog


logger = logging.getLogger(__name__)


class Nem(QObject):

    def __init__(self, app):
        super(Nem, self).__init__(parent=app)
        self._app = app
        self.login_dialog = LoginDialog(
            verify_captcha=LoginController.check_captcha,
            verify_userpw=LoginController.check,
            create_user=LoginController.create,
        )
        self._user = None
        self._pm = None

    def ready_to_login(self):
        if self._user is not None:
            logger.debug('You have already logined in.')
            return
        logger.debug('Trying to load last login user...')
        user = LoginController.load()
        if user is None or 'MUSIC_U' not in user.cookies:
            logger.debug('Trying to load last login user...failed')
            self.login_dialog.show()
            self.login_dialog.load_user_pw()
            self.login_dialog.login_success.connect(
                lambda user: asyncio.ensure_future(self.login_as(user)))
        else:
            logger.debug('Trying to load last login user...done')
            asyncio.ensure_future(self.login_as(user))

    async def login_as(self, user):
        provider.auth(user)
        self._user = user
        LoginController.save(user)
        left_panel = self._app.ui.left_panel
        loop = asyncio.get_event_loop()
        self._pm.name = '网易云音乐 - {}'.format(user.name)
        playlists = await loop.run_in_executor(None, lambda: user.playlists)
        self._app.playlists.add(playlists)
        self._app.playlists.add(user.fav_playlists, is_fav=True)



#! /usr/bin/env python3

import argparse
import asyncio
import logging
import os

import traceback
import sys

from fuocore.app import run_server
from fuocore.pubsub import run as run_pubsub

from feeluown import logger_config
from feeluown.rcfile import load_rcfile
from feeluown.consts import (
    HOME_DIR, USER_PLUGINS_DIR, PLUGINS_DIR, DATA_DIR,
    CACHE_DIR, USER_THEMES_DIR, SONG_DIR
)
from feeluown.utils import is_port_used

logger = logging.getLogger(__name__)


def ensure_dir():
    if not os.path.exists(HOME_DIR):
        os.mkdir(HOME_DIR)
    if not os.path.exists(DATA_DIR):
        os.mkdir(DATA_DIR)
    if not os.path.exists(USER_PLUGINS_DIR):
        os.mkdir(USER_PLUGINS_DIR)
    if not os.path.exists(CACHE_DIR):
        os.mkdir(CACHE_DIR)
    if not os.path.exists(USER_THEMES_DIR):
        os.mkdir(USER_THEMES_DIR)
    if not os.path.exists(SONG_DIR):
        os.mkdir(SONG_DIR)


def excepthook(exc_type, exc_value, tb):
    """Exception hook"""
    src_tb = tb
    while src_tb.tb_next:
        src_tb = src_tb.tb_next
    logger.error('Exception occured: print locals varibales')
    logger.error(src_tb.tb_frame.f_locals)
    traceback.print_exception(exc_type, exc_value, tb)


ensure_dir()
os.chdir(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
sys.path.append(PLUGINS_DIR)
sys.path.append(USER_PLUGINS_DIR)


def setup_argparse():
    parser = argparse.ArgumentParser(description='运行 FeelUOwn 播放器')
    parser.add_argument('-nw', '--no-window', action='store_true', default=False,
                        help='以 CLI 模式运行')

    parser.add_argument('-d', '--debug', action='store_true', default=False,
                        help='开启调试模式')
    parser.add_argument('--log-to-file', action='store_true', default=False,
                        help='将日志打到文件中')

    # XXX: 不知道能否加一个基于 regex 的 option？比如加一个
    # `--mpv-*` 的 option，否则每个 mpv 配置我都需要写一个 option？

    # TODO: 需要在文档中给出如何查看有哪些播放设备的方法
    parser.add_argument(
        '--mpv-audio-device',
        default='auto',
        help='（高级选项）给 mpv 播放器指定播放设备'
    )
    return parser


def main():
    parser = setup_argparse()
    args = parser.parse_args()

    if is_port_used(23333) or is_port_used(23334):
        print('\033[0;31m', end='')
        print('Port(23333 or 23334) is used, maybe another feeluown is running?')
        print('\033[0m', end='')
        sys.exit(1)

    debug = args.debug
    mpv_audio_device = args.mpv_audio_device
    cli_only = args.no_window
    logger_config(debug, to_file=args.log_to_file)

    from fuocore.player import MpvPlayer

    player = MpvPlayer(audio_device=bytes(mpv_audio_device, 'utf-8'))
    player.initialize()

    # 设置 exception hook
    sys.excepthook = excepthook

    if not cli_only:
        try:
            import PyQt5  # noqa
        except ImportError:
            logger.warning('PyQt5 is not installed，can only use CLI mode.')
            cli_only = True

    if not cli_only:
        from PyQt5.QtWidgets import QApplication
        from quamash import QEventLoop
        from feeluown.guiapp import GuiApp

        q_app = QApplication(sys.argv)
        q_app.setQuitOnLastWindowClosed(True)
        q_app.setApplicationName('FeelUOwn')

        app_event_loop = QEventLoop(q_app)
        asyncio.set_event_loop(app_event_loop)
        pubsub_gateway, pubsub_server = run_pubsub()

        app = GuiApp(pubsub_gateway, player=player)
        app.initialize()
        load_rcfile(app)
        # TODO: 调用 show 时，会弹出主界面，但这时界面还没开始绘制
        # 为了让提升启动速度，一些非必须的初始化操作可以在 show 之后进行
        app.show()
    else:
        from feeluown.app import CliApp

        pubsub_gateway, pubsub_server = run_pubsub()
        app = CliApp(pubsub_gateway)
        app.initialize()

    live_lyric = app.live_lyric
    event_loop = asyncio.get_event_loop()
    event_loop.create_task(run_server(app, live_lyric))
    try:
        event_loop.run_forever()
        logger.info('Event loop stopped.')
    except KeyboardInterrupt:
        # NOTE: gracefully shutdown?
        pass
    finally:
        pubsub_server.close()
        event_loop.close()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import os
import importlib
import logging

from .consts import USER_PLUGINS_DIR, PLUGINS_DIR


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class PluginsManager(object):
    def __init__(self, app):
        super().__init__()
        self._app = app

        self._plugins = {}

    def load(self, plugin):
        plugin.enable(self._app)

    def unload(self, plugin):
        plugin.disable(self._app)

    def scan(self):
        logger.debug('Scaning plugins...')
        modules_name = [p for p in os.listdir(PLUGINS_DIR)
                        if os.path.isdir(os.path.join(PLUGINS_DIR, p))]
        modules_name.extend([p for p in os.listdir(USER_PLUGINS_DIR)
                            if os.path.isdir(os.path.join(USER_PLUGINS_DIR))])
        for module_name in modules_name:
            try:
                module = importlib.import_module(module_name)
                plugin_name = module.__alias__
                self._plugins[plugin_name] = module
                module.enable(self._app)
                logger.info('detect plugin: %s.' % plugin_name)
            except:
                logger.exception('detect a bad plugin %s' % module_name)
        logger.debug('Scaning plugins...done')



import platform
import socket


def parse_ms(ms):
    minute = int(ms / 60000)
    second = int((ms % 60000) / 1000)
    return minute, second


def is_linux():
    if platform.system() == 'Linux':
        return True
    return False


def is_osx():
    if platform.system() == 'Darwin':
        return True
    return False


def emit_requests_progress(response, signal=None):
    content = bytes()
    total_size = response.headers.get('content-length')
    if total_size is None:
        content = response.content
        return content
    else:
        total_size = int(total_size)
        bytes_so_far = 0

        for chunk in response.iter_content(102400):
            content += chunk
            bytes_so_far += len(chunk)
            progress = round(bytes_so_far * 1.0 / total_size * 100)
            if signal is not None:
                signal.emit(progress)
        return content


def is_port_used(port, host='0.0.0.0'):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    rv = sock.connect_ex((host, port))
    return rv == 0



import asyncio
import logging
import os
from contextlib import contextmanager
from functools import partial

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QImage, QPixmap, QIcon
from PyQt5.QtWidgets import (
    QApplication,
    QStyle,
    QWidget,
)

from fuocore.player import State as PlayerState

from feeluown.components.history import HistoriesModel
from feeluown.components.provider import ProvidersModel
from feeluown.components.playlists import PlaylistsModel

from .app import CliApp, App
from .consts import APP_ICON
from .helpers import use_mac_theme
from .hotkey import Hotkey
from .img_ctl import ImgController
from .request import Request
from .tips import TipsManager
from .ui import Ui

logger = logging.getLogger(__name__)


class AppCodeRunnerMixin(object):
    def __init__(self):
        self._g = {}

    def exec_(self, code):
        obj = compile(code, '<string>', 'single')
        self._g.update({
            'app': self,
            'player': self.player
        })
        exec(obj, self._g, self._g)


class GuiApp(CliApp, QWidget, AppCodeRunnerMixin):
    """
    FIXME: Subclass must call initialize to make this mixin
    work properly, which seems to be little bit strange. But
    this mixin helps avoid duplicate code temporarily.
    """

    mode = App.GuiMode | App.CliMode

    def __init__(self, pubsub_gateway, player=None):
        CliApp.__init__(self, pubsub_gateway, player=None)
        QWidget.__init__(self)

        self.request = Request(self)
        self.tips_manager = TipsManager(self)
        self.hotkey_manager = Hotkey(self)
        self.img_ctl = ImgController(self)

        self.playlists = PlaylistsModel(parent=self)
        self.histories = HistoriesModel(parent=self)
        self.providers = ProvidersModel(parent=self)

        self.ui = Ui(self)

        self.player_pixmap = None

        self.resize(1000, 618)
        self.setObjectName('app')
        QApplication.setWindowIcon(QIcon(APP_ICON))

        self.bind_signal()

    def initialize(self):
        super().initialize()
        self.load_qss()
        self.tips_manager.show_random_tip()

    def load_qss(self):
        if not use_mac_theme():
            return
        filepath = os.path.abspath(__file__)
        dirname = os.path.dirname(filepath)
        qssfilepath = os.path.join(dirname, 'mac.qss')
        with open(qssfilepath) as f:
            s = f.read()

            QApplication.instance().setStyleSheet(s)

    def bind_signal(self):
        top_panel = self.ui.top_panel

        self.player.state_changed.connect(self._on_player_status_changed)
        self.player.position_changed.connect(self._on_player_position_changed)
        self.player.duration_changed.connect(self._on_player_duration_changed)
        # FIXME:
        self.player.playlist.playback_mode_changed.connect(
            top_panel.pc_panel.on_playback_mode_changed)
        self.player.playlist.song_changed.connect(
            top_panel.pc_panel.on_player_song_changed)

        self.request.connected_signal.connect(self._on_network_connected)
        self.request.disconnected_signal.connect(self._on_network_disconnected)
        self.request.slow_signal.connect(self._on_network_slow)
        self.request.server_error_signal.connect(self._on_network_server_error)

        #top_panel.pc_panel.volume_slider.sliderMoved.connect(
        #    self.change_volume)

    @contextmanager
    def create_action(self, s):
        show_msg = self.ui.magicbox.show_msg

        class Action(object):
            def set_progress(self, value):
                value = int(value * 100)
                show_msg(s + '...{}%'.format(value), timeout=-1)

            def failed(self):
                show_msg(s + '...failed', timeout=-1)

        show_msg(s + '...', timeout=-1)  # doing
        try:
            yield Action()
        except Exception:
            show_msg(s + '...error')  # error
            raise
        else:
            show_msg(s + '...done')  # done

    def _on_player_position_changed(self, ms):
        self.ui.top_panel.pc_panel.on_position_changed(ms*1000)
        self.ui.top_panel.pc_panel.progress_slider.update_state(ms*1000)

    def _on_player_duration_changed(self, ms):
        self.ui.top_panel.pc_panel.on_duration_changed(ms*1000)
        self.ui.top_panel.pc_panel.progress_slider.set_duration(ms*1000)

    def _on_player_status_changed(self, state):
        pp_btn = self.ui.top_panel.pc_panel.pp_btn
        if use_mac_theme():
            return
        if state == PlayerState.playing:
            pp_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaPause))
        else:
            pp_btn.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))

    def _on_network_slow(self):
        pass

    def _on_network_connected(self):
        pass

    def _on_network_server_error(self):
        pass

    def _on_network_disconnected(self):
        pass

    def change_volume(self, value):
        self.player.volume = value

    def pixmap_from_url(self, url, callback=None):
        # FIXME: only neteasemusic img url accept the params
        data = {'param': '{0}y{0}'.format(self.width())}
        res = self.request.get(url, data)
        if res is None:
            return None
        img = QImage()
        img.loadFromData(res.content)
        pixmap = QPixmap(img)
        if pixmap.isNull():
            return None
        if callback is not None:
            callback(pixmap)
        return pixmap

    def closeEvent(self, event):
        try:
            self.player.stop()
            self.player.shutdown()
        except Exception as e:
            pass
        QApplication.quit()

    def resizeEvent(self, event):
        super().resizeEvent(event)



# -*- coding: utf-8 -*-

import logging

from fuocore.qqmusic.provider import provider
from feeluown.app import App

__alias__ = 'QQ 音乐'
__feeluown_version__ = '1.1.0'
__version__ = '0.1a0'
__desc__ = 'QQ 音乐'

logger = logging.getLogger(__name__)


def enable(app):
    app.library.register(provider)
    if app.mode & App.GuiMode:
        from feeluown.components.provider import ProviderModel

        pm = ProviderModel(
            name='QQ 音乐',
            icon='♫ ',
            desc='点击登录 QQ 音乐（未实现，欢迎 PR）',
            on_click=None,
        )
        app.providers.assoc(provider.identifier, pm)


def disable(app):
    app.library.deregister(provider)
    if app.mode & App.GuiMode:
        app.providers.remove(provider.identifier)



from PyQt5.QtWidgets import QFrame


class Separator(QFrame):
    def __init__(self, orientation='horizontal', parent=None):
        super().__init__(parent)

        if orientation == 'horizontal':
            self.setFrameShape(QFrame.HLine)
        else:
            self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Sunken)



# -*- coding: utf-8 -*-

import logging

from fuocore.netease.provider import provider



__alias__ = '网易云音乐'
__feeluown_version__ = '1.0.4.2'
__version__ = '0.0.2'
__desc__ = '网易云音乐'

logger = logging.getLogger(__name__)


def enable(app):
    app.library.register(provider)
    if app.mode & app.GuiMode:
        from feeluown.components.provider import ProviderModel
        from .nem import Nem

        nem = Nem(app)
        pm = ProviderModel(
            name='网易云音乐',
            desc='点击可以登录',
            on_click=nem.ready_to_login,
        )
        nem._pm = pm
        app.providers.assoc(provider.identifier, pm)


def disable(app):
    app.library.deregister(provider)
    if app.mode & app.GuiMode:
        app.provider.remove(provider.identifier)



import io
import sys

from PyQt5.QtCore import (
    pyqtSignal,
    QSize,
    Qt,
    QRect,
    QTimer,
)
from PyQt5.QtGui import (
    QColor,
    QFont,
    QTextCursor,
    QFontDatabase,
    QFontMetrics,
    QPainter,
    QPalette,
    QSyntaxHighlighter,
    QTextCharFormat,
    QTextOption,
)
from PyQt5.QtWidgets import (
    QApplication,
    QLineEdit,
    QStyle,
    QStyleOptionFrame,
    QTextEdit,
)


class Highlighter(QSyntaxHighlighter):
    """
    Python REPL 和 fuo 命令语法高亮
    """
    def __init__(self, parent=None):
        super().__init__(parent)

        prompt_fmt = QTextCharFormat()
        prompt_fmt.setFontWeight(QFont.Bold)
        self._prompt_fmt = prompt_fmt

    def highlightBlock(self, text):
        if text.startswith('>>> '):
            self.setFormat(0, 3, self._prompt_fmt)


class MagicBox(QTextEdit):
    """读取用户输入，解析执行

    ref: https://wiki.qt.io/Technical_FAQ #How can I create a one-line QTextEdit?
    """

    returnPressed = pyqtSignal()

    def __init__(self, app, parent=None):
        super().__init__(parent)

        self._app = app
        # self.setPlaceholderText('搜索歌曲、歌手、专辑、用户；执行 Python 代码等')
        self.setPlaceholderText('Search library, exec code, or run command.')
        self.setToolTip('直接输入文字可以进行过滤，按 Enter 可以搜索\n'
                        '输入 >>> 前缀之后，可以执行 Python 代码\n'
                        '输入 $ 前缀可以执行 fuo 命令（未实现，欢迎 PR）')
        self.setFont(QFontDatabase.systemFont(QFontDatabase.FixedFont))

        # 模拟 QLineEdit
        self.setTabChangesFocus(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setWordWrapMode(QTextOption.NoWrap)
        self.setFixedHeight(self.sizeHint().height())
        self.setMinimumHeight(25)

        self._timer = QTimer(self)
        self._cmd_text = None
        self._mode = 'cmd'  # 详见 _set_mode 函数
        self._timer.timeout.connect(self.__on_timeout)

        self.textChanged.connect(self.__on_text_edited)
        self._highlighter = Highlighter(self.document())
        # self.textEdited.connect(self.__on_text_edited)
        self.returnPressed.connect(self.__on_return_pressed)

    def show_msg(self, text, timeout=2000):
        if not text:
            return
        self._set_mode('msg')
        self.setText(text)
        if timeout > 0:
            self._timer.start(timeout)

    def _set_mode(self, mode):
        """修改当前模式

        现在主要有两种模式：cmd 模式是正常模式；msg 模式用来展示消息通知，
        当自己处于 msg 模式下时，会 block 所有 signal
        """
        if mode == 'cmd':
            self.setReadOnly(False)
            self._timer.stop()
            if self._cmd_text:
                self.setHtml(self._cmd_text)
            # 注意在所有操作完成之后再关闭 blockSignals
            # 然后修改当前 mode
            self.blockSignals(False)
            self._mode = mode
        elif mode == 'msg':
            self.blockSignals(True)
            if self._mode == 'cmd':
                self.setReadOnly(True)
                self._cmd_text = self.toHtml()
                self._mode = mode

    def _search_library(self, q):
        songs = []
        for result in self._app.library.search(q):
            songs.extend(result.songs)
        self._app.ui.table_container.show_songs(songs)

    def _exec_code(self, code):
        """执行代码并重定向代码的 stdout/stderr"""
        output = io.StringIO()
        sys.stderr = output
        sys.stdout = output
        try:
            self._app.exec_(code)
        except Exception as e:
            print(str(e))
        finally:
            sys.stderr = sys.__stderr__
            sys.stdout = sys.__stdout__
        self.show_msg(output.getvalue() or 'No output.')

    def __on_text_edited(self):
        text = self.toPlainText()
        if self._mode == 'cmd':
            self._cmd_text = self.toHtml()
        if not text.startswith('>'):
            self._app.ui.table_container.search(text)

    def __on_return_pressed(self):
        text = self.toPlainText()
        if text.startswith('>>> '):
            self._exec_code(text[4:])
        else:
            self._search_library(text)

    def __on_timeout(self):
        self._set_mode('cmd')

    def focusInEvent(self, e):
        super().focusInEvent(e)
        self._set_mode('cmd')

    def keyPressEvent(self, e):
        if e.key() in (Qt.Key_Return, Qt.Key_Enter):
            self.returnPressed.emit()
            e.ignore()
        else:
            super().keyPressEvent(e)

    def wheelEvent(self, e):
        if self._mode != 'cmd':
            super().wheelEvent(e)

    def sizeHint(self):
        fm = QFontMetrics(self.font())
        # FIXME: 暂时不是太懂应该怎样正确的计算 w 和 h
        # 其中，计算 h 的一个原则是让高度正好能够显示一行
        # 目前在 Linux 下测试 4 正好符合期望
        h = max(fm.height() + fm.ascent() - fm.descent(), 14) + 4
        w = self.width() - 4
        opt = QStyleOptionFrame()
        opt.initFrom(self)
        return self.style().sizeFromContents(
            QStyle.CT_LineEdit,
            opt,
            QSize(w, h).expandedTo(QApplication.globalStrut()),
            self
        )



import logging

from PyQt5.QtCore import (
    pyqtSignal,
    QAbstractListModel,
    Qt,
    QModelIndex,
    QPoint,
    QRect,
    QSize,
    QTimer,
    QVariant,
)
from PyQt5.QtGui import (
    QBrush,
    QColor,
    QPainter,
    QPen,
    QFontMetrics,
)
from PyQt5.QtWidgets import (
    QAbstractItemDelegate,
    QAbstractItemView,
    QComboBox,
    QHBoxLayout,
    QHeaderView,
    QItemDelegate,
    QListView,
    QPushButton,
    QSizePolicy,
    QStyledItemDelegate,
    QSpinBox,
    QTableView,
    QTableWidget,
    QWidget,
)

from feeluown.helpers import action_log, ActionError


logger = logging.getLogger(__name__)


class PlaylistsModel(QAbstractListModel):
    def __init__(self, parent):
        super().__init__(parent)
        self._playlists = []
        self._fav_playlists = []

    def add(self, playlist, is_fav=False):
        if is_fav:
            start = len(self._playlists) + len(self._fav_playlists)
            playlists = self._fav_playlists
        else:
            start = len(self._playlists)
            playlists = self._playlists

        if isinstance(playlist, list):
            _playlists = playlist
        else:
            _playlists = [playlist]
        end = start + len(_playlists)
        self.beginInsertRows(QModelIndex(), start, end)
        playlists.extend(playlist)
        self.endInsertRows()

    def rowCount(self, _):
        return len(self._playlists) + len(self._fav_playlists)

    def flags(self, index):
        if not index.isValid():
            return 0
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled
        if index.row() < len(self._playlists):
            flags |= Qt.ItemIsDropEnabled
        return flags

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        # XXX: 实际不产生任何效果
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return '播放列表'
        return QVariant()

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return QVariant()
        row = index.row()
        # 新创建的 playlists 中的元素是引用
        playlists = self._playlists + self._fav_playlists
        if row >= len(playlists) or row < 0:
            return QVariant()

        playlist = playlists[row]
        if role == Qt.DisplayRole:
            if row < len(self._playlists):
                flag = '♬ '
            else:
                flag = '★ '
            return flag + playlist.name
        elif role == Qt.UserRole:
            return playlist
        return QVariant()


class PlaylistsView(QListView):

    show_playlist = pyqtSignal([object])

    def __init__(self, parent):
        super().__init__(parent)

        self.setAttribute(Qt.WA_MacShowFocusRect, 0)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)

        self.setDragDropMode(QAbstractItemView.DropOnly)
        self.clicked.connect(self._on_clicked)

        self._result_timer = QTimer(self)
        self._result_timer.timeout.connect(self.__on_timeout)
        self._results = {}  # {row: [index, True]}

    def _on_clicked(self, index):
        playlist = index.data(role=Qt.UserRole)
        self.show_playlist.emit(playlist)

    def __on_timeout(self):
        self._result_timer.stop()
        self._results.clear()
        self.viewport().update()

    def dropEvent(self, e):
        mimedata = e.mimeData()
        if mimedata.hasFormat('fuo-model/x-song'):
            song = mimedata.model
            index = self.indexAt(e.pos())
            playlist = index.data(Qt.UserRole)
            if song.source != playlist.source:
                e.ignore()
                return
            with action_log('Add {} to {}'.format(song, playlist)):
                self._results[index.row] = (index, None)
                self.viewport().update()
                is_success = playlist.add(song.identifier)
                self._results[index.row] = (index, is_success)
                self.viewport().update()
                self._result_timer.start(2000)
                if not is_success:
                    raise ActionError
            e.accept()
        else:
            e.ignore()

    def paintEvent(self, e):
        super().paintEvent(e)
        if not self._results:
            return
        painter = QPainter(self.viewport())
        option = self.viewOptions()
        painter.setRenderHint(QPainter.Antialiasing)
        fm = QFontMetrics(option.font)
        for row, result in self._results.items():
            index, state = result
            rect = self.rectForIndex(index)
            if state is None:
                text = '⇵'
            elif state is True:
                text = '👋'
            else:
                text = '🙁'
            x = rect.width() - 20 + rect.x()
            # 让字垂直居中
            y = (rect.height() + fm.ascent() - fm.descent()) / 2 + rect.y()
            topleft = QPoint(x, y)
            painter.drawText(topleft, text)

    def dragEnterEvent(self, e):
        e.accept()

    def dragMoveEvent(self, e):
        index = self.indexAt(e.pos())
        if index.flags() & Qt.ItemIsDropEnabled:
            e.accept()
        else:
            e.ignore()



from PyQt5.QtCore import (
    QAbstractTableModel,
    Qt,
    QTime,
    QVariant,
)
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QHeaderView,
    QTableWidget,
    QTableWidgetItem,
    QTableView,
)


class SongsTable(QTableView):

    def __init__(self, parent=None):
        super().__init__(parent)

        # self.setEditTriggers(QAbstractItemView.NoEditTriggers)
        # self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.horizontalHeader().setDefaultAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        # self.verticalHeader().hide()
        # self.setShowGrid(False)
        # self.setAlternatingRowColors(True)
        # self.setColumnWidth(0, 28)
        # self.setColumnWidth(2, 150)
        # self.setColumnWidth(3, 150)
        # self.setColumnWidth(4, 50)
        # self.setColumnWidth(5, 30)
        # self.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        # self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        # self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    # def keyPressEvent(self, event):
    #     self.setFocus()     # gain focus from cell widget if neccesary
    #     key_code = event.key()
    #     if key_code == Qt.Key_J:
    #         self.setCurrentCell(self._next_row(), 1)
    #     elif key_code == Qt.Key_K:
    #         self.setCurrentCell(self._prev_row(), 1)
    #     elif key_code in (Qt.Key_Enter, Qt.Key_Return):
    #         current_row = self.currentRow()
    #     else:
    #         super().keyPressEvent(event)

    def _next_row(self):
        current_row = self.currentRow()
        return current_row + 1 if current_row != (self.rowCount() - 1)\
            else current_row

    def _prev_row(self):
        current_row = self.currentRow()
        return current_row - 1 if current_row != 0 else 0



from collections import deque

from PyQt5.QtCore import (
    pyqtSignal,
    QAbstractListModel,
    QModelIndex,
    QSize,
    Qt,
    QVariant,
)
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QListView,
    QSizePolicy,
)

from fuocore import ModelType


class HistoriesModel(QAbstractListModel):
    def __init__(self, size=5, parent=None):
        super().__init__(parent)
        self._size = size
        self._models = deque()

    def append(self, model):
        # TODO: change strategy to LRU
        curlen = len(self._models)
        if model in self._models:
            self.beginInsertRows(QModelIndex(), curlen - 2, curlen - 1)
            self._models.remove(model)
            self._models.append(model)
            self.endInsertRows()
            return

        if curlen > self._size:
            self.beginInsertRows(QModelIndex(), curlen - 1, curlen)
            self._models.popleft()
        else:
            self.beginInsertRows(QModelIndex(), curlen, curlen + 1)
        self._models.append(model)
        self.endInsertRows()

    def rowCount(self, parent=QModelIndex()):
        return len(self._models)

    def flags(self, index):
        return Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return QVariant()
        row = index.row()
        if row >= len(self._models) or row < 0:
            return QVariant()

        # latest first
        model = self._models[len(self._models) - row - 1]
        if role == Qt.DisplayRole:
            model_type = ModelType(model._meta.model_type)
            if model_type == ModelType.song:
                return model.title
            elif model_type in (ModelType.playlist,
                                ModelType.artist,
                                ModelType.album):
                return model.name
            else:
                return str(model)
        elif role == Qt.UserRole:
            return model
        return QVariant()


class HistoriesView(QListView):
    """显示最近浏览的播放列表、歌手、专辑等"""

    show_model = pyqtSignal([object])

    def __init__(self, parent):
        super().__init__(parent)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setSelectionMode(QAbstractItemView.NoSelection)
        self.setAttribute(Qt.WA_MacShowFocusRect, 0)

        self.clicked.connect(self._on_clicked)

    def _on_clicked(self, index):
        model = index.data(role=Qt.UserRole)
        self.show_model.emit(model)

    def sizeHint(self):
        return QSize(self.width(), 100)

    def currentChanged(self, current, previous):
        """让最近浏览的元素流畅的显示在列表最上方"""
        first = self.model().index(0, 0)
        super().currentChanged(first, previous)



# -*- coding: utf-8 -*-
import locale
import logging

from fuocore.player import MpvPlayer


logger = logging.getLogger(__name__)


class Player(MpvPlayer):

    def __init__(self):
        locale.setlocale(locale.LC_NUMERIC, 'C')
        super().__init__()
        self.initialize()



import os


DEFAULT_RCFILE_PATH = os.path.expanduser('~/.fuorc')


def load_rcfile(app, rcfile_path=DEFAULT_RCFILE_PATH):
    if not os.path.exists(DEFAULT_RCFILE_PATH):
        return

    g = {'app': app}
    with open(rcfile_path) as f:
        exec(f.read(), g)



import hashlib
import json
import logging
import os

from PyQt5.QtCore import pyqtSignal, Qt, pyqtSlot, QTime
from PyQt5.QtGui import QColor, QImage, QPixmap
from PyQt5.QtWidgets import (QHBoxLayout, QVBoxLayout, QLineEdit, QHeaderView,
                             QMenu, QAction, QAbstractItemView,
                             QTableWidgetItem, QSizePolicy, QDialog, QFrame, QPushButton,
                             QScrollArea, QLabel)

from .consts import USER_PW_FILE

logger = logging.getLogger(__name__)


class LoginDialog(QDialog):
    login_success = pyqtSignal([object])

    def __init__(self, verify_captcha, verify_userpw, create_user, parent=None):
        super().__init__(parent)

        self.verify_captcha = verify_captcha
        self.verify_userpw = verify_userpw
        self.create_user = create_user

        self.is_encrypted = False
        self.captcha_needed = False
        self.captcha_id = 0

        self.username_input = QLineEdit(self)
        self.pw_input = QLineEdit(self)
        self.pw_input.setEchoMode(QLineEdit.Password)
        # self.remember_checkbox = FCheckBox(self)
        self.captcha_label = QLabel(self)
        self.captcha_label.hide()
        self.captcha_input = QLineEdit(self)
        self.captcha_input.hide()
        self.hint_label = QLabel(self)
        self.ok_btn = QPushButton('登录', self)
        self._layout = QVBoxLayout(self)

        self.username_input.setPlaceholderText('网易邮箱或者手机号')
        self.pw_input.setPlaceholderText('密码')

        self.pw_input.textChanged.connect(self.dis_encrypt)
        self.ok_btn.clicked.connect(self.login)

        self.setFixedWidth(200)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.username_input)
        self._layout.addWidget(self.pw_input)
        self._layout.addWidget(self.captcha_label)
        self._layout.addWidget(self.captcha_input)
        self._layout.addWidget(self.hint_label)
        # self._layout.addWidget(self.remember_checkbox)
        self._layout.addWidget(self.ok_btn)

    def fill(self, data):
        self.username_input.setText(data['username'])
        self.pw_input.setText(data['password'])
        self.is_encrypted = True

    def show_hint(self, text):
        self.hint_label.setText(text)

    @property
    def data(self):
        username = self.username_input.text()
        pw = self.pw_input.text()
        if self.is_encrypted:
            password = pw
        else:
            password = hashlib.md5(pw.encode('utf-8')).hexdigest()
        d = dict(username=username, password=password)
        return d

    def captcha_verify(self, data):
        self.captcha_needed = True
        url = data['captcha_url']
        self.captcha_id = data['captcha_id']
        self.captcha_input.show()
        self.captcha_label.show()
        # FIXME: get pixmap from url
        # self._app.pixmap_from_url(url, self.captcha_label.setPixmap)

    def dis_encrypt(self, text):
        self.is_encrypted = False

    def login(self):
        if self.captcha_needed:
            captcha = str(self.captcha_input.text())
            captcha_id = self.captcha_id
            data = self.check_captcha(captcha_id, captcha)
            if data['code'] == 200:
                self.captcha_input.hide()
                self.captcha_label.hide()
            else:
                self.captcha_verify(data)

        user_data = self.data
        self.show_hint('正在登录...')
        data = self.verify_userpw(user_data['username'], user_data['password'])
        message = data['message']
        self.show_hint(message)
        if data['code'] == 200:
            self.save_user_pw(user_data)
            user = self.create_user(data)
            self.login_success.emit(user)
            self.hide()
        elif data['code'] == 415:
            self.captcha_verify(data)

    def save_user_pw(self, data):
        with open(USER_PW_FILE, 'w+') as f:
            if f.read() == '':
                d = dict()
            else:
                d = json.load(f)
            d['default'] = data['username']
            d[d['default']] = data
            json.dump(d, f, indent=4)

        logger.info('save username and password to %s' % USER_PW_FILE)

    def load_user_pw(self):
        if not os.path.exists(USER_PW_FILE):
            return
        with open(USER_PW_FILE, 'r') as f:
            d = json.load(f)
            data = d[d['default']]
        self.username_input.setText(data['username'])
        self.pw_input.setText(data['password'])
        self.is_encrypted = True

        logger.info('load username and password from %s' % USER_PW_FILE)



import asyncio
import logging

from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QImage, QPixmap, QFont, QPalette, QPainter, QColor
from PyQt5.QtWidgets import (
    QDialog,
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListView,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QTabBar,
    QVBoxLayout,
    QWidget,
)

from fuocore import ModelType
from feeluown.helpers import use_mac_theme
from feeluown.components.songs_table import SongsTableModel, SongsTableView


logger = logging.getLogger(__name__)


class DescriptionContainer(QScrollArea):
    def __init__(self, parent=None):
        super().__init__(parent)

        self._label = QLabel(self)
        self._label.setWordWrap(True)
        self._label.setTextFormat(Qt.RichText)
        self._label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.setWidget(self._label)
        self.setWidgetResizable(True)

        self.setFrameShape(QFrame.NoFrame)
        self._label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

    @property
    def html(self):
        return self._label.text()

    def set_html(self, desc):
        self._label.setText(desc)

    def keyPressEvent(self, event):
        key_code = event.key()
        if key_code == Qt.Key_J:
            value = self.verticalScrollBar().value()
            self.verticalScrollBar().setValue(value + 20)
        elif key_code == Qt.Key_K:
            value = self.verticalScrollBar().value()
            self.verticalScrollBar().setValue(value - 20)
        else:
            super().keyPressEvent(event)


class TableOverview(QFrame):
    def __init__(self, parent):
        super().__init__(parent)

        self._height = 180
        self.cover_label = QLabel(self)
        self._desc_container = DescriptionContainer(self)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)
        self._layout = QHBoxLayout(self)
        self._left_sub_layout = QVBoxLayout()
        self._right_sub_layout = QVBoxLayout()

        self._right_sub_layout.addWidget(self._desc_container)
        self._left_sub_layout.addWidget(self.cover_label)
        self._left_sub_layout.addStretch(0)
        self._layout.addLayout(self._left_sub_layout)
        self._layout.addSpacing(20)
        self._layout.addLayout(self._right_sub_layout)
        self._layout.setStretch(1, 1)
        self.cover_label.setFixedWidth(200)
        self.setFixedHeight(self._height)

    def set_cover(self, pixmap):
        self.cover_label.setPixmap(
            pixmap.scaledToWidth(self.cover_label.width(),
                                 mode=Qt.SmoothTransformation))

    def set_desc(self, desc):
        self._desc_container.show()
        self._desc_container.set_html(desc)

    def keyPressEvent(self, event):
        key_code = event.key()
        if key_code == Qt.Key_Space:
            if self._height < 300:
                self._height = 300
                self.setMinimumHeight(self._height)
                self.setMaximumHeight(self._height)
            else:
                self._height = 180
                self.setMinimumHeight(self._height)
                self.setMaximumHeight(self._height)
            event.accept()


class SongsTableContainer(QFrame):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self.songs_table = SongsTableView(self)
        self.table_overview = TableOverview(self)

        self._layout = QVBoxLayout(self)

        self.setAutoFillBackground(False)
        if use_mac_theme():
            self._layout.setContentsMargins(0, 0, 0, 0)
            self._layout.setSpacing(0)
        self._layout.addWidget(self.table_overview)
        self._layout.addWidget(self.songs_table)

        self.songs_table.play_song_needed.connect(
            lambda song: asyncio.ensure_future(self.play_song(song)))
        self.songs_table.show_artist_needed.connect(
            lambda artist: asyncio.ensure_future(self.show_model(artist)))
        self.songs_table.show_album_needed.connect(
            lambda album: asyncio.ensure_future(self.show_model(album)))
        self._cover_pixmap = None
        self.hide()

    async def play_song(self, song):
        loop = asyncio.get_event_loop()
        with self._app.create_action('play {}'.format(song)):
            await loop.run_in_executor(None, lambda: song.url)
            self._app.player.play_song(song)

    def play_all(self):
        songs = self.songs_table.model().songs
        self._app.player.playlist.clear()
        for song in songs:
            self._app.player.playlist.add(song)
        self._app.player.play_next()

    async def show_model(self, model):
        model_type = ModelType(model._meta.model_type)
        if model_type == ModelType.album:
            func = self.show_album
        elif model_type == ModelType.artist:
            func = self.show_artist
        elif model_type == ModelType.playlist:
            func = self.show_playlist
        else:
            def func(model): pass  # seems silly
        self._app.histories.append(model)
        with self._app.create_action('show {}'.format(str(model))):
            await func(model)

    def show_player_playlist(self, songs):
        self.show_songs(songs)
        self.songs_table.song_deleted.connect(
            lambda song: self._app.playlist.remove(song))

    async def show_playlist(self, playlist):
        self.table_overview.show()
        loop = asyncio.get_event_loop()
        songs = await loop.run_in_executor(None, lambda: playlist.songs)
        self._show_songs(songs)
        desc = '<h2>{}</h2>\n{}'.format(playlist.name, playlist.desc or '')
        self.table_overview.set_desc(desc)
        if playlist.cover:
            loop.create_task(self.show_cover(playlist.cover))

        def remove_song(song):
            model = self.songs_table.model()
            row = model.songs.index(song)
            # 如果有 f-string 该有多好！
            msg = 'remove {} from {}'.format(song, playlist)
            with self._app.create_action(msg) as action:
                rv = playlist.remove(song.identifier)
                if rv:
                    model.removeRow(row)
                else:
                    action.failed()

        self.songs_table.song_deleted.connect(lambda song: remove_song(song))

    async def show_artist(self, artist):
        self.table_overview.show()
        loop = asyncio.get_event_loop()
        future_songs = loop.run_in_executor(None, lambda: artist.songs)
        future_desc = loop.run_in_executor(None, lambda: artist.desc)
        await asyncio.wait([future_songs, future_desc])
        desc = future_desc.result()
        self.table_overview.set_desc(desc or '<h2>{}</h2>'.format(artist.name))
        self._show_songs(future_songs.result())
        if artist.cover:
            loop.create_task(self.show_cover(artist.cover))

    async def show_album(self, album):
        loop = asyncio.get_event_loop()
        future_songs = loop.run_in_executor(None, lambda: album.songs)
        future_desc = loop.run_in_executor(None, lambda: album.desc)
        await asyncio.wait([future_songs, future_desc])
        self.table_overview.set_desc(future_desc.result() or
                                     '<h2>{}</h2>'.format(album.name))
        songs = future_songs.result()
        self._show_songs(songs)
        if album.cover:
            loop.create_task(self.show_cover(album.cover))

    async def show_cover(self, cover):
        # FIXME: cover_hash may not work properly someday
        cover_uid = cover.split('/', -1)[-1]
        content = await self._app.img_ctl.get(cover, cover_uid)
        img = QImage()
        img.loadFromData(content)
        pixmap = QPixmap(img)
        if not pixmap.isNull():
            self.table_overview.set_cover(pixmap)
            self.update()

    def _show_songs(self, songs):
        try:
            self.songs_table.song_deleted.disconnect()
        except TypeError:  # no connections at all
            pass
        self.show()
        songs = songs or []
        logger.debug('Show songs in table, total: %d' % len(songs))
        source_name_map = {p.identifier: p.name for p in self._app.library.list()}
        self.songs_table.setModel(SongsTableModel(songs, source_name_map))
        self.songs_table.scrollToTop()

    def show_songs(self, songs):
        self._show_songs(songs)
        self.table_overview.hide()

    def search(self, text):
        if self.isVisible() and self.songs_table is not None:
            self.songs_table.filter_row(text)



# -*- coding: utf-8 -*-

import os
from enum import Enum


THEMES_DIR = './feeluown/themes'
PLUGINS_DIR = './feeluown/plugins'
APP_ICON = './feeluown/feeluown.png'
DEFAULT_THEME_NAME = 'Tomorrow Night'
DEFAULT_THEME_NAME = 'Solarized'


HOME_DIR = os.path.expanduser('~') + '/.FeelUOwn'

DATA_DIR = HOME_DIR + '/data'
USER_PLUGINS_DIR = HOME_DIR + '/plugins'
USER_THEMES_DIR = HOME_DIR + '/themes'
CACHE_DIR = HOME_DIR + '/cache'
SONG_DIR = HOME_DIR + '/songs'

LOG_FILE = HOME_DIR + '/stdout.log'


class PlaybackMode(Enum):
    one_loop = '🔂'
    loop = '🔁'
    random = '🔀'



import asyncio
import locale
import logging
from functools import partial

from fuocore import LiveLyric, MpvPlayer, Library
from fuocore.app import LiveLyricPublisher

from .plugin import PluginsManager
from .version import VersionManager


logger = logging.getLogger(__name__)


class App(object):
    CliMode = 0x0001
    GuiMode = 0x0010

    mode = 0x0000

    def __init__(self, player=None):
        if player is None:
            locale.setlocale(locale.LC_NUMERIC, 'C')
            self.player = MpvPlayer()
            self.player.initialize()
        else:
            self.player = player
        self.playlist = self.player.playlist
        self.library = Library()
        self.live_lyric = LiveLyric()

        self.plugin_mgr = PluginsManager(self)
        self.version_mgr = VersionManager(self)

    def initialize(self):
        self.player.position_changed.connect(self.live_lyric.on_position_changed)
        self.playlist.song_changed.connect(self.live_lyric.on_song_changed)

        self.plugin_mgr.scan()
        loop = asyncio.get_event_loop()
        loop.call_later(10, partial(loop.create_task, self.version_mgr.check_release()))


class CliApp(App):
    mode = App.CliMode

    def __init__(self, pubsub_gateway, player=None):
        super().__init__(player=player)

        self.pubsub_gateway = pubsub_gateway
        self._live_lyric_publisher = LiveLyricPublisher(pubsub_gateway)

        self.live_lyric.sentence_changed.connect(self._live_lyric_publisher.publish)
